import ByteIO
import StringIO			
import sys
import Qid			
import User			
import Fid				
import Errors			

OREAD  = 0x00
OWRITE = 0x01
ORDWR  = 0x02
OEXEC  = 0x03
OTRUNC = 0x10
OCLOSE = 0x40

BIT16SZ = 2

MAXWELEM = 16 # apparently

headersize = 4 + 1 + 2
authenticated = 0
debug = 0

def hex_format(data) :
	counter = ''
	dec = ''
	chrs = ''
	ex = ''
		
	for i in xrange(0, len(data)) :
		c = ord(data[i])
		if c < 32 or c > 128 :
			ch = 32
		else :
			ch = c
		counter = counter + "%03d " % (i + headersize)
		dec = dec + "%03d " % (c)
		chrs = chrs + "%3c " % (ch)
		h = hex(c)[1:]
		h = (3-len(h)) * ' ' + h
		ex = ex + "%s " % (h)
			
	return counter + '\n' + dec + '\n' + chrs + '\n' + ex + '\n'

	
# ============================= T Message ============================================

class Message:
	
	def __init__(self, tag=0, size=0) :
		self.tag = tag
		self.size = size
		global debug
		if debug :
			Errors.stderr("New : %s" % (self.name))
	
	def __str__(self) :
		bytes = self.body_bytes().getvalue()
		counter = ''
		dec = ''
		chrs = ''
		ex = ''
		
		for i in xrange(0, len(bytes)) :
			c = ord(bytes[i])
			if c < 32 or c > 128 :
				ch = 32
			else :
				ch = c
			counter = counter + "%03d " % (i + headersize)
			dec = dec + "%03d " % (c)
			chrs = chrs + "%3c " % (ch)
			h = hex(c)[1:]
			h = (3-len(h)) * ' ' + h
			ex = ex + "%s " % (h)
			
		return self.name + '\n' + counter + '\n' + dec + '\n' + chrs + '\n' + ex + '\n'
		
	def load(self, size, bytes) :
		self.load_from_bytes(bytes)
		
		if bytes.io_bytes_read != size :
			raise 'wrong sized message', " exp: %d   got: %d" % (size, bytes.io_bytes_read)
			
	def write_to(self, stdout) :
		global headersize
		body = self.body_bytes().getvalue()
		bio = ByteIO.ByteIO(stdout)
		bio.PBIT32(len(body) + headersize)
		bio.PBIT8(self.type)
		bio.PBIT16(self.tag)
		bio.write(body)
		try :
			stdout.flush()
		except :
			pass
		
	def __getattr__(self, attr) :
		if attr == 'name' :
			try :
				str = Message_classes[self.type][1]
			except :
				str = 'Unknown type'
			return str
		else :
			raise AttributeError, self.name + "." + attr 

class Tmessage(Message) :
		
	def load_from_bytes(self, bytes) :
		raise 'Tmessage.load_from_bytes - base class not overridden - bad news'

	def get_Rmessage(self) :
		raise ("Tmessage - get_Rmessage not overridden type %d" % (self.type,))
	
		
class Tversion(Tmessage) :
	type = 100
		
	def load_from_bytes(self, bytes) :
		self.msize = bytes.GBIT32()
		self.version = bytes.gstring()
		if self.version[0:2] == '9P' :
			self.version = '9P2000'
		else :
			self.version = 'unknown'
			
	def body_bytes(self) :
		bio = ByteIO.ByteIO(StringIO.StringIO())
		
		bio.PBIT32(self.msize)
		bio.pstring(self.version)
		
		return bio

	def get_Rmessage(self) :
		r = Rversion(self.type + 1, self.tag)
		r.msize = self.msize
		r.version = self.version

		return r

class Tflush(Tmessage) :
	type = 108	

	def load_from_bytes(self, bytes) :
		self.oldtag = bytes.GBIT16()

	def body_bytes(self) :
		bio = ByteIO.ByteIO(StringIO.StringIO())
		bio.PBIT16(self.oldtag)
		return bio

	def get_Rmessage(self) :
		r = Rflush(self.type + 1, self.tag)
		return r

class Tauth(Tmessage) :
	type = 102
	
	def load_from_bytes(self, bytes) :
		self.afid = bytes.GBIT32()
		self.uname = bytes.gstring()
		self.aname = bytes.gstring()
		
	def body_bytes(self) :
		bio = ByteIO.ByteIO(StringIO.StringIO())
		
		bio.PBIT32(self.afid)
		bio.pstring(self.uname)
		bio.pstring(self.aname)
		
		return bio

	def get_Rmessage(self) :
		r = Rauth(self.type + 1, self.tag)
		r.uname = self.uname
		r.aname = self.aname

		global authenticated
		authenticated = 1
		return r

class Tattach(Tmessage) :	
	type = 104
		
	def load_from_bytes(self, bytes) :
		self.fid = bytes.GBIT32()
		self.afid = bytes.GBIT32()
		self.uname = bytes.gstring()
		self.aname = bytes.gstring()
		if self.aname == 'device' :
			raise 'Tattach aname was set to "device" which was marked as a special case'
					
	def body_bytes(self) :
		bio = ByteIO.ByteIO(StringIO.StringIO())
		bio.PBIT32(self.fid)
		bio.PBIT32(self.afid)
		bio.pstring(self.uname)
		bio.pstring(self.aname)
		return bio

	def get_Rmessage(self) :
		r = Rattach(self.type + 1, self.tag)

		r.fid = self.fid
		r.afid = self.afid
		r.uname = self.uname
		r.aname = self.aname

		return r


class Twalk(Tmessage) :	
	type = 110
	
	def __init__(self, tag=0, size=0) :
		
		Tmessage.__init__(self, tag)
		
	def load_from_bytes(self, bytes) :
		self.fid = bytes.GBIT32()
		self.newfid = bytes.GBIT32()
		self.nwname = bytes.GBIT16()
		self.wname = {}


		if self.nwname > MAXWELEM :
			raise "Twalk nwname > MAXWELEM %d" % (MAXWELEM)

		for i in xrange(0, self.nwname) :
			self.wname[i] = bytes.gstring()

		if self.nwname == 0 :
			self.wname[0] = '/'
			self.nwname = 1
			
	def body_bytes(self) :
		bio = ByteIO.ByteIO(StringIO.StringIO())
		bio.PBIT32(self.fid)
		bio.PBIT32(self.newfid)
		try :
			nwname = len(self.wname)
			if nwname > 0 :
				bio.PBIT16(nwname)
				for wname in self.wname :
					bio.pstring(wname)
			else :
				bio.PBIT16(0)
		except :
			bio.PBIT16(0)
			
		return bio
			

	def get_Rmessage(self) :
		r = Rwalk(self.type + 1, self.tag)
		
		r.fid = self.fid
		r.newfid = self.newfid
		r.wname = self.wname

		return r




class Topen(Tmessage) :	
	type = 112
		
	def load_from_bytes(self, bytes) :
		self.fid = bytes.GBIT32()
		self.mode = bytes.GBIT8()

	def body_bytes(self) :
		bio = ByteIO.ByteIO(StringIO.StringIO())
		bio.PBIT32(self.fid)
		bio.PBIT8(self.mode)
	
		return bio
	
	
	
class Tcreate(Tmessage) :	
	type = 114
		
	def load_from_bytes(self, bytes) :
		self.fid =  bytes.GBIT32()
		self.name =  bytes.gstring()
		self.perm =  bytes.GBIT32()
		self.mode =  bytes.GBIT8()

class Tread(Tmessage) :
	type = 116
	
	def body_bytes(self) :
		bio = ByteIO.ByteIO(StringIO.StringIO())
		bio.PBIT32(self.fid)
		bio.PBIT64(self.offset)
		bio.PBIT32(self.count)
	
		return bio
		
		
	def load_from_bytes(self, bytes) :
		self.fid = bytes.GBIT32()
		self.offset = bytes.GBIT64()
		self.count = bytes.GBIT32()

class Twrite(Tmessage) :	
	type = 118
		
	def load_from_bytes(self, bytes) :
		self.fid = bytes.GBIT32()
		self.offset = bytes.GBIT64()
		self.count = bytes.GBIT32()
		self.data = bytes.read(self.count)

class Tclunk(Tmessage) :				
	type = 120
		
	def load_from_bytes(self, bytes) :
		self.fid = bytes.GBIT32()
		
	def body_bytes(self) :
		bio = ByteIO.ByteIO(StringIO.StringIO())
		bio.PBIT32(self.fid)
	
		return bio

	def get_Rmessage(self) :
		r = Rclunk(self.type + 1, self.tag)
		r.fid = self.fid

		return r

class Tremove(Tmessage) :		
	type = 122
		
	def load_from_bytes(self, bytes) :
		self.fid = bytes.GBIT32()	

class Tstat(Tmessage) :				
	type = 124
		
	def load_from_bytes(self, bytes) :
		self.fid = bytes.GBIT32()
		
	def body_bytes(self) :
		bio = ByteIO.ByteIO(StringIO.StringIO())
		bio.PBIT32(self.fid)
		return bio

	def get_Rmessage(self) :
		r = Rstat(self.type + 1, self.tag)
		r.fid = self.fid

		return r


class Twstat(Tmessage) :		
	type = 126
		
	def load_from_bytes(self, bytes) :
		self.fid = bytes.GBUIT32()
		self.stat = bytes.gstring()
		self.nstat = len(self.stat)
		
class Terror :	
	type = 106
		
	def load_from_bytes(self, bytes) :
		raise "Terror"

# ============================= R Message ============================================
class Rmessage(Message) :
	pass
	
class Rversion(Rmessage) :		
	type = 101
		
	def body_bytes(self) :
		bio = ByteIO.ByteIO(StringIO.StringIO())
		
		bio.PBIT32(self.msize)
		bio.pstring(self.version)
		
		return bio
		
	def load_from_bytes(self, bytes) :
		self.msize = bytes.GBIT32()
		self.version = bytes.gstring()
		if self.version[0:2] == '9P' :
			self.version = '9P2000'
		else :
			self.version = 'unknown'

class Rauth(Rmessage) :	
	type = 103

	def body_bytes(self) :
		self.type = Rerror.type
		bio = ByteIO.ByteIO(StringIO.StringIO())
		bio.pstring('no authentication required')

		return bio

class Rattach(Rmessage) :
	type = 105
		
	def body_bytes(self) :
		
		fid = Fid.newfid(self.fid)
		fid.path = '/'

		if not fid.stat() :
			Fid.freefid(fid)
			raise 'Rattach fidstat failed for /'

		u = User.User()
		u.getpwnam(self.uname)
		u._print()

		if not u.uid > 0: #  no roots here thanks
			Fid.freefid(fid)
			raise ("Rattach unknown user %s" %(self.uname,))

		fid.set_u(u)
		fid._print()

		qid = Qid.Qid()
		
		qid.make_from_fid_stat(fid.st)

		qid._print()

		bio = ByteIO.ByteIO(StringIO.StringIO())
		bio.pqid(qid)

		return bio

	def load_from_bytes(self, bytes) :
		self.qid = Qid.Qid()
		self.qid.type = bytes.GBIT8()
		self.qid.vers = bytes.GBIT32()
		self.qid.path = bytes.GBIT64()

class Rwalk(Rmessage) :
	type = 111

	def body_bytes(self) :
		self.wqid = []

		fid = Fid.oldfid(self.fid)
		if fid.omode != -1 :
			raise Errors.Ebadusefid

		if not fid.stat() :
			raise 'Rwalk - oldfid stat failed'

		if not fid.isdir() and self.nwname < 1 :
			raise 'Rwalk - fid a file and no wname supplied'

		if self.newfid != self.fid :
			nfid = Fid.newfid(self.newfid)
			if not nfid :
				raise 'Rwalk - newfid failed'
		
		path = fid.path

		qid = None
		lastqid = None
		for bit in self.wname :
			path = os.path.join(path, self.wname[bit])
			qid = Qid.Qid(path)
			if not qid.path :
				break
			self.wqid.append(qid)
			
			lastqid = qid

		if qid == None :
			if lastqid == None :
				raise 'Twalk failed'

		bio = ByteIO.ByteIO(StringIO.StringIO())
		for wname in self.wname :
			bio.pstring(self.wname[wname])

		return bio
		
	def load_from_bytes(self, bytes) :
		nwqid = bytes.GBIT16()
		self.wqid = []
		for i in xrange(0, nwqid) :
			self.wqid.append(bytes.gqid())
	
	
class Rclunk(Rmessage) :
	type = 121
	
	def load_from_bytes(self, bytes) :
		pass

	def body_bytes(self) :
		f = Fid.oldfid(self.fid, -1)
		if not f :
			raise 'Rclunk - Fid.oldfid failed'

		if f.auth :
			raise "Rclunk - auth was set but i don't auth"

		if f.omode != -1 and f.omode | ORCLOSE :
			os.unlink(fid.path)

		Fid.freefid(f)

		bio = ByteIO.ByteIO(StringIO.StringIO())
		bio.PBIT32(self.fid)

		return bio

class Rstat(Rmessage) :
	type = 125
	
	def body_bytes(self) :
		f = Fid.oldfid(self.fid)
		if not f :
			raise 'Rstat - oldfid failed'
		if not f.stat() :
			raise 'Rstat - stat failed'

		dir = f.stat2dir()
		if not dir :
			raise 'Rstat - stat2dir failed'

		# convD2M
		d2m = dir.D2M()
		if len(dir.D2M()) <= BIT16SZ :
			raise 'Rstat - dir.S2M failed'
		bio = ByteIO.ByteIOO(StringIO.StringIO(d2m))
		return bio

	def load_from_bytes(self, bytes) :
		self.stat_size = bytes.GBIT16()
		self.st_size = bytes.GBIT16()
		self.st_type = bytes.GBIT16()
		self.st_dev = bytes.GBIT32()
		self.st_qid = bytes.gqid()
		self.st_mode = bytes.GBIT32()
		self.st_atime = bytes.GBIT32()
		self.st_mtime = bytes.GBIT32()
		self.st_length = bytes.GBIT64()
		self.st_name = bytes.gstring()
		if self.st_name == '' :
			self.st_name = '/'
		self.st_uid = bytes.gstring()
		self.st_gid = bytes.gstring()
		self.st_muid = bytes.gstring()
		
	def load_from_stat_bytes(self, bytes) :
		self.st_size = bytes.GBIT16()
		self.st_type = bytes.GBIT16()
		self.st_dev = bytes.GBIT32()
		self.st_qid = bytes.gqid()
		self.st_mode = bytes.GBIT32()
		self.st_atime = bytes.GBIT32()
		self.st_mtime = bytes.GBIT32()
		self.st_length = bytes.GBIT64()
		self.st_name = bytes.gstring()
		if self.st_name == '' :
			self.st_name = '/'
		self.st_uid = bytes.gstring()
		self.st_gid = bytes.gstring()
		self.st_muid = bytes.gstring()		
		
class Rerror(Rmessage) :
	type = 107
		
	def load_from_bytes(self, bytes) :
		self.ename =  bytes.gstring()
		
	def body_bytes(self) :
		bio = ByteIO.ByteIO(StringIO.StringIO())
		bio.pstring(self.ename)

		return bio
	
class Rflush(Rmessage) :
	type = 109
	
	def body_bytes(self) :
		bio = ByteIO.ByteIO(StringIO.StringIO())
		return bio

class Ropen(Rmessage) :
	type = 113
			
	def load_from_bytes(self, bytes) :
		self.qid = bytes.gqid()
		self.iounit = bytes.GBIT32()
	
class Rcreate(Rmessage) :
	type = 115
	
class Rread(Rmessage) :
	type = 117
	
	def load_from_bytes(self, bytes) :
		global headersize
		self.count = bytes.GBIT32()
		# ignore count because if it is result of statting a directory it is not the same format		
		self.data = bytes.GDATA(self.size - bytes.io_bytes_read)
	
class Rwrite(Rmessage) :
	type = 119
	
class Rremove(Rmessage) :
	type = 123
		
class Rwstat(Rmessage) :
	type = 127
	
	
	
	
Message_classes = { 
	  0 : (None, 'None'),
	100 : (Tversion, 'Tversion'), 
	101 : (Rversion, "Rversion"),
	102 : (Tauth, "Tauth"), 
	103: (Rauth, "Rauth"), 
	104 : (Tattach, "Tattach"), 
	105 : (Rattach, "Rattach"),
	106 : (Terror, "Terror"), 
	107 : (Rerror, "Rerror"), 
	108 : (Tflush, "Tflush"), 
	109 : (Rflush, "Rflush"), 
	110 : (Twalk, "Twalk"), 
	111 : (Rwalk, "Rwalk"), 
	112 : (Topen, "Topen"), 
	113 : (Ropen, "Ropen"), 
	114 : (Tcreate, "Tcreate"), 
	115 : (Rcreate, "Rcreate"), 
	116 : (Tread, "Tread"), 
	117 : (Rread, "Rread"), 
	118 : (Twrite, "Twrite"), 
	119 : (Rwrite, "Rwrite"), 
	120 : (Tclunk, "Tclunk"), 
	121 : (Rclunk, "Rclunk"), 
	122 : (Tremove, "Tremove"), 
	123 : (Rremove, "Rremove"), 
	124 : (Tstat, "Tstat"), 
	125 : (Rstat, "Rstat"), 
	126 : (Twstat, "Twstat"), 
	127 : (Rwstat , "Rwstat"),
	128 : (None , "Tmax")
}

def read_file(block_size, tag, fid, channel, offset=0) :
	''' read the contents of the file starting at offset'''
	offset = int(offset)
	count = int(block_size)
	data = ''
	
	t = Tread(tag)
	t.fid = fid
	t.offset = offset
	t.count = count
	
	t.write_to(channel)
	r = get_message(channel)
	
	try :
		while len(r.data) > 0 :
			data = data + r.data
			offset = offset + count
			count = count + block_size
			t = Tread(tag)
			t.fid = fid
			t.offset = offset
			t.count = count
	
			t.write_to(channel)
			r = get_message(channel)
	except :
		pass
		
	return data
		


# get a message from an instance
# bytes  must have read/write methods
# and for Tmessages bytes needs getvalue

def get_message(fin) : # fin needs a read(n)
	bytes = ByteIO.ByteIO(fin)
	m = None
	if bytes :
		size = bytes.GBIT32()
		type = bytes.GBIT8()
		tag = bytes.GBIT16()
		m = Message_classes[type][0](tag, size)
		if m :
			if 1 : #try :
				m.load(size, bytes)
			#except :
			#	raise 'load failed'		
		else :
			raise 'unknown message type', type		
	return m
